#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <signal.h>
#include <sys/types.h>
#include <glib.h>
#include <glib/gstdio.h>
#include "config.h"
#include "serial.h"
#include "splash.h"
#include "session.h"
#include "mouse.h"
#include "telmicom.h"

#ifdef __linux__
#      define PATH_PROGRAM "/cyber"
#      define SCREENSAVER "kdesktop_lock"
#      define EVENT
#      define UNUSED  __attribute__((__unused__))
#	include <unistd.h>
#elif __WIN32__
#      define PATH_PROGRAM "C:/WINDOWS/system32/config/cyber"
#      define EVENT __declspec(dllexport)
#      define UNUSED
#      include <windows.h>
#      include <ws2tcpip.h>
#endif
#ifdef __linux__
	#define msleep(x) usleep(1000*(x))
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <netdb.h>
	#include <arpa/inet.h>
	#include <sys/wait.h>
#elif __WIN32__
	#define msleep(x) Sleep(x)
	#include <ws2tcpip.h>
	void WSAAPI freeaddrinfo (struct addrinfo*);
	int WSAAPI getaddrinfo (const char*,const char*,const struct addrinfo*, struct addrinfo**);
	int WSAAPI getnameinfo(const struct sockaddr*,socklen_t,char*,DWORD, char*,DWORD,int);
	#define close(x) closesocket(x)
	typedef int socklen_t;
#endif

#define FILE_CONFIG     (PATH_PROGRAM "/cyber.conf")
#define FILE_GLADE   PATH_PROGRAM "/data/resta_credit.glade"
#define MESSAGE_MAX	2000
#define LISTEN_PORT "7777"

typedef enum
{
	esperant, iniciat, cancelat, acceptat, cobrat
} job_status_t;

typedef struct
{
	serial_t *serial;
	GArray *missatges;
	unsigned char org;
	unsigned char dst;
	double preu;
	int temps;
	int idiomes;
	int lang;
	int print_pages;
	double preu_pagina;
	int printer_timeout;
	job_status_t job_status;
	int job_init;
} status_t;

status_t status = {
	NULL,
	NULL,
	1,
	2,
	3.0,
	0,
	4,
	0,
	0,
	.3,
	30,
	esperant,
	0,
};

GKeyFile *config = NULL;

char text_timer[10];
char *messages[4];
static GtkWidget *window;
char ** messages_ok;
char ** messages_ko;
int idiomes;
int lang=0;
GKeyFile *f;
char message[MESSAGE_MAX + 1];
GladeXML *mainXML = NULL;
int sockfd= -1;
FILE *loggerfile = NULL;

void logger(const char *format, ...)

{
	va_list ap;
	GTimeVal t;

	va_start(ap, format);
	if (loggerfile) {
		g_get_current_time(&t);
		fprintf(loggerfile, "%s: ", g_time_val_to_iso8601(&t));
		vfprintf(loggerfile, format, ap);
		fprintf(loggerfile, "\n");
		fflush(loggerfile);

		g_printf("%s: ", g_time_val_to_iso8601(&t));
		g_vprintf(format, ap);
		g_printf("\n");
	}

	va_end(ap);
}

void subst_label(char* src, char *label, char* value)
{
//	char format[16];
	char buff[MESSAGE_MAX + 1];
	char *p, *q;
	int i = 0,len = strlen(label), len_val = strlen(value);
	p = src;
	while ((q = strstr(p,label)) && i < MESSAGE_MAX) {
		*q =  0;
		strncpy(buff + i, p, MESSAGE_MAX - i);
		i += q - p;
		strncpy(buff + i, value, MESSAGE_MAX - i);
		i += len_val;
		p = q + len;
	}
	strncpy(buff + i, p, MESSAGE_MAX - i);
	strncpy(src, buff, MESSAGE_MAX);
}

void gen_message()
{
	double missing = 0, total_price;
	char buff[32];
	int segons, remain_time;
	if (lang < 0 || lang >= idiomes) return;

	total_price = ((double)status.print_pages) * status.preu_pagina;

	segons = total_price * 3600.0 / status.preu;
	remain_time = status.temps - segons;
	if ( status.temps < segons ) {
		missing = ((double) (segons - status.temps )) * status.preu / 3600.0;
		strncpy(message, messages_ko[lang], MESSAGE_MAX);
		gtk_widget_hide(glade_xml_get_widget(mainXML,"button_ok"));
	} else {
		strncpy(message, messages_ok[lang], MESSAGE_MAX);
		gtk_widget_show(glade_xml_get_widget(mainXML,"button_ok"));
	}

	snprintf(buff, 30, "%d", status.print_pages);
	subst_label(message, "[NUM_PAG]", buff);

	snprintf(buff, 30, "%01.2f", status.preu_pagina);
	subst_label(message, "[PRICE_PAG]", buff);

	snprintf(buff, 30, "%01.2f", total_price);
	subst_label(message, "[TOTAL_COST]", buff);

	snprintf(buff, 30, "%02d:%02d:%02d", segons / 3600, (segons/60)%60, segons % 60);
	subst_label(message, "[TOTAL_TIME]", buff);

	snprintf(buff, 30, "%02d:%02d:%02d", status.temps / 3600, (status.temps/60)%60, status.temps % 60);
	subst_label(message, "[AVAIL_TIME]", buff);

	snprintf(buff, 30, "%02d:%02d:%02d", remain_time / 3600, (remain_time/60)%60, remain_time % 60);
	subst_label(message, "[REMAIN_TIME]", buff);

	snprintf(buff, 30, "%01.2f", status.preu);
	subst_label(message, "[PRICE_HOUR]", buff);

	snprintf(buff, 30, "%01.2f", missing);
	subst_label(message, "[MISSING_CASH]", buff);
}

EVENT gboolean lang_changed(GtkWidget *widget UNUSED, GdkEvent*event UNUSED, gpointer user_data)
{
	int t = ((int)user_data) - 1;
	if ( t >= 0 && t < idiomes) lang= t;
	return TRUE;
}

void afegir_bandera(GtkHBox * lang_container, int index)
{
	GtkWidget * eventBox = gtk_event_box_new();
	GtkWidget * image;
	char buff[128];
	gtk_widget_show(eventBox);
	snprintf(buff,125,"%s/data/idioma_%d.gif",PATH_PROGRAM,index);
	gtk_box_pack_start(GTK_BOX(lang_container),eventBox,0,0,5);
	image = gtk_image_new_from_file(buff);
	gtk_widget_show(image);
    gtk_container_add(GTK_CONTAINER(eventBox),image);
	gtk_widget_show(image);
	gtk_widget_show(eventBox);

	g_signal_connect ((gpointer) eventBox, "button_press_event", G_CALLBACK (lang_changed), (gpointer)index);
}

void createWindow() {
	int c;
	if (window) return;
	if (mainXML) free(mainXML);
	mainXML = glade_xml_new(FILE_GLADE, NULL, NULL);
	glade_xml_signal_autoconnect(mainXML);

	GtkHBox * lang_container = (GtkHBox *)glade_xml_get_widget(mainXML,"lang_container");
	for (c = 1 ; c <= idiomes ; c++) afegir_bandera(lang_container, c);

	window = (GtkWidget *)glade_xml_get_widget(mainXML,"dialog1");
//	gtk_signal_connect (GTK_OBJECT (window), "destroy", gtk_widget_destroyed, &window);
	gtk_widget_show(window);
}

gboolean redraw_timer(gpointer data UNUSED)
{
	static job_status_t old_status = esperant;
//	printf ("Dins Redraw Timer\n");
	if (status.job_status != esperant) {
		createWindow();
		gen_message();
		GtkLabel * label = (GtkLabel *) glade_xml_get_widget(mainXML,"label1");
		gtk_label_set_markup(label,message);
		if (old_status == esperant) gtk_widget_show(window);
		//gtk_widget_queue_resize(GTK_WIDGET(glade_xml_get_widget(mainXML,"dialog1")));
		gtk_window_deiconify(GTK_WIDGET(glade_xml_get_widget(mainXML,"dialog1")));
		gtk_window_set_keep_above(GTK_WINDOW(glade_xml_get_widget(mainXML,"dialog1")), TRUE);
	} else {
		if (old_status != esperant) gtk_widget_hide((GtkWidget *)glade_xml_get_widget(mainXML,"dialog1"));
	}
	old_status = status.job_status;

	return TRUE;
}


gpointer printer_loop(gpointer data)
{
	// Inicialitzar socket

	int sockfd, new_fd; // listen on sock_fd, new connection on new_fd
	int sockfd2, new_fd2; // listen on sock_fd, new connection on new_fd
	struct addrinfo hints, *servinfo, *p;
	struct sockaddr_storage their_addr; // connector's address information
	socklen_t sin_size;
	char s[INET6_ADDRSTRLEN];
	int rv;
	char buff[128];
	int i;
	int yes = 1;
	GTimeVal now;
	//daemon_config_t *printer_config = (daemon_config_t *) data;

	logger("printer_loop:init");

#ifdef __WIN32__
	WSADATA wsaData;
	WSAStartup(MAKEWORD(2,0),&wsaData);
#endif
	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE; // use my IP

	if ((rv = getaddrinfo(NULL, LISTEN_PORT, &hints, &servinfo)) == -1) {
		return 1;
	}

	// loop through all the results and bind to the first we can
	for (p = servinfo; p != NULL; p = p->ai_next) {
		if ((sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol))
		        == -1) {
			logger("Error: printer_loop : server: socket");
			continue;
		}

		if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int))
		        == -1) {
			logger("Error: printer_loop : setsockopt");
			exit(1);
		}

		if (bind(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
			close(sockfd);
			logger("Error: printer_loop : server: bind");
			continue;
		}

		break;
	}

	if (p == NULL) {
		logger("Error: printer_loop : server: failed to bind\n");
		return 2;
	}

	freeaddrinfo(servinfo); // all done with this structure

	if (listen(sockfd, 10) == -1) {
		logger("Error: printer_loop : listen");
		exit(1);
	}

	logger("printer_loop: waiting for connections...\n");

	while (1) { // main accept() loop
		int id, pagines, preu_pagina, c;
		char md5[100];
		char *sign;
		sin_size = sizeof their_addr;
		new_fd = accept(sockfd, (struct sockaddr *) &their_addr, &sin_size);
		if (new_fd == -1) {
			logger("Error: printer_loop : accept");
			continue;
		}

		logger("printer_loop: got connection\n");

		if (recv(new_fd, buff, 100, 0) > 0) {
			buff[100] = 0;
			if (sscanf(buff, "%d,%d,%d,%s", &id, &pagines, &preu_pagina, md5)
			        < 4) goto PRINTER_SOCKET_ERROR;
			sign = printer_sign(KEY_SENDPAGES, id, pagines, preu_pagina);
			if (strcmp(md5, sign)) {
				logger("Invalid signature!");
				goto PRINTER_SOCKET_ERROR;
			}

			if (pagines <= 0) goto PRINTER_SOCKET_ERROR;

			//			if (!status.ui_init) goto PRINTER_SOCKET_ERROR;
			g_get_current_time(&now);
			status.print_pages = pagines;
			status.preu_pagina = ((double) preu_pagina) / 100.0;
			status.job_init = now.tv_sec;
			status.job_status = iniciat;

			for (; status.job_status != cobrat && status.job_status != cancelat; msleep(500)) {
				g_get_current_time(&now);
				if (status.job_status == iniciat) {
					if ((now.tv_sec - status.job_init)
					        > status.printer_timeout) {
						logger("print timeout: %d %d %d", now.tv_sec,
						        status.job_init,
						        status.printer_timeout);
						snprintf(buff, 100, "%d\n", -1);
						send(new_fd2, buff, strlen(buff), 0);
						status.job_status = cancelat;
					}
				}
			}
			close(new_fd2);
			close(sockfd2);

			if (status.job_status == cobrat) {
				i = 1;
			} else i = 0;
			snprintf(buff, 100, "%d,%d\n", id, i);
			send(new_fd, buff, strlen(buff), 0);
		}
		PRINTER_SOCKET_ERROR: status.job_status = esperant;
		close(new_fd);
	}

}

gpointer telmicom_loop(gpointer data)
{
	int lag, t;
	double tarifa;
	int i = 0;

	for (;;) {
		t = telmicom_temps(status.serial, status.org, status.dst);
//		logger(" **** Llegit temps: %d \n", t);
		if (t >= 0) status.temps = t;
		if (i % 500 == 0) {
			i=0;
//			logger(" **** dins if: %d \n", t);
			tarifa = telmicom_get_tarifa(status.serial, status.org, status.dst);
			if (tarifa > 0) status.preu = tarifa;
			telmicom_descuento(status.serial, status.org, status.dst, 1);
		}
		i++;
		if (status.job_status == acceptat) {
			int descompte_segons = (int) (((double) status.print_pages) * status.preu_pagina * 3600.0 / status.preu);
			if (descompte_segons > 0 && (lag = status.temps - descompte_segons) > -2) {
//				logger("Descompant segons %d - %d = %d %02d:%02d:%02d", status.temps, descompte_segons, lag, lag / 3600, (lag / 60) % 60, lag % 60);
				if (!telmicom_settemps(status.serial, status.org, status.dst, lag > 0 ? lag : 0))
					status.job_status = cobrat;
			}
		}

	}
	return NULL;
}

int config_load(const char *filename)
{
	GError *error;
	GKeyFile *f;

	error = NULL;
	f = g_key_file_new();
	if (!g_key_file_load_from_file(f, filename, G_KEY_FILE_KEEP_COMMENTS,
	        &error)) {
		g_print("Error: Can't open the configuration file '%s'.", filename);
		return -1;
	}

	if (config) {
		g_key_file_free(config);
	}

	config = f;

	return 0;
}

serial_t *init_serial()
{
	GError *error;
	GThread *thread;
	serial_config_t *sc;
	serial_t *serial;
	char *parity;
	char *dev;
	int port;
	gboolean isUsb = FALSE;

	logger("Starting serial configuration.");

	if (!(sc = serial_config_create())) {
		logger("Error: Can't configure the Serial Port.");

		return NULL;
	}

	error = NULL;
	port = g_key_file_get_integer(config, "com", "port", &error);

	error = NULL;
	isUsb = g_key_file_get_boolean(config, "com", "usb", &error);

	error = NULL;
	sc->speed = g_key_file_get_integer(config, "com", "speed", &error);

	error = NULL;
	sc->word_size = g_key_file_get_integer(config, "com", "bits", &error);

	error = NULL;
	sc->stop_bits = g_key_file_get_integer(config, "com", "stopbits", &error);

	error = NULL;
	parity = g_key_file_get_string(config, "com", "parity", &error);

	if (!parity || !*parity) {
		logger("Device: No parity defined");

		return NULL;
	}

	switch (*parity) {
		case 'N':
		case 'n':
			sc->parity = SERIAL_PARITY_NONE;
			break;

		case 'E':
		case 'e':
			sc->parity = SERIAL_PARITY_EVEN;
			break;

		case 'O':
		case 'o':
			sc->parity = SERIAL_PARITY_ODD;
			break;

		case 'M':
		case 'm':
			sc->parity = SERIAL_PARITY_MARK;
			break;

		case 'S':
		case 's':
			sc->parity = SERIAL_PARITY_SPACE;
			break;

		default:
			sc->parity = SERIAL_PARITY_NONE;
	}

	sc->parity = 0;

	logger("Device: com=%d; speed=%d; w=%d; p=%d; s=%d;", port, sc->speed,
	        sc->word_size, sc->parity, sc->stop_bits);

	if (!(serial = serial_init())) {
		logger("Device: Can't init serial.");
		serial_config_release(sc);

		return NULL;
	}

	if (!(dev = serial_device(port, isUsb))) {
		logger("Device: Can't open serial device. Unknown device.");
		free(dev);
	}

	if (serial_open(serial, dev, sc) < 0) {
		logger("Device: Can't open serial device.");
		serial_config_release(sc);
		free(dev);

		return NULL;
	}

	serial_config_release(sc);
	free(dev);

	return serial;
}


int main(int argc, char **argv)
{
//	GtkWidget *image;
	GError *error;
	int c;
//	char *b;
	char buff[32];

	gtk_init(&argc, &argv);

	f = g_key_file_new();
	error = NULL;
	if (!g_key_file_load_from_file(f, FILE_CONFIG, G_KEY_FILE_KEEP_COMMENTS, &error)) {
		printf("Error: Can't find file '%s'\n", FILE_CONFIG);
		return -1;
	}

	error = 0;
	idiomes = g_key_file_get_integer(f, "printer_msg", "idiomes", &error);

	messages_ok = (char **) malloc (sizeof(char * )*idiomes);
	messages_ko = (char **) malloc (sizeof(char * )*idiomes);

	for (c = 0; c < idiomes; c++) {
		sprintf(buff,"msg_ok_%d",c);
		error = 0;
		messages_ok[c] = g_key_file_get_string(f, "printer_msg", buff, &error);
		sprintf(buff,"msg_ko_%d",c);
		error = 0;
		messages_ko[c] = g_key_file_get_string(f, "printer_msg", buff, &error);
		printf("messages[%d]=%s | %s;\n", c, messages_ok[c], messages_ko[c]);
	}

	g_key_file_free(f);

	if (!(loggerfile = fopen(FILE_LOG, "w"))) {
		g_print("Error: Can't open '%s' for writing\n", FILE_LOG);

		return 1;
	}
	if (config_load(FILE_CONFIG) < 0) {
		return 2;
	}
	for (;;) {
		status.serial = init_serial();
		if (status.serial) break;
		logger("No s'ha pogut obrir el port sèrie. Reintentant");
		msleep(10000);
	}

	g_timeout_add(100, redraw_timer, NULL);
	g_thread_init(NULL);
	g_thread_create(printer_loop, NULL, TRUE, &error);
	g_thread_create(telmicom_loop, NULL, TRUE, &error);




	gtk_main();

	return 0;
}

EVENT gboolean destroy(GtkWidget *widget UNUSED, gpointer user_data UNUSED)
{
	status.job_status = cancelat;
	window = NULL;
}

EVENT gboolean ok_clicked(GtkWidget *widget UNUSED, gpointer user_data UNUSED)
{
	status.job_status = acceptat;
}


EVENT gboolean cancel_clicked(GtkWidget *widget UNUSED, gpointer user_data UNUSED)
{
	status.job_status = cancelat;
}
