#ifndef __SERIAL_H__
#define __SERIAL_H__

#define SERIAL_PARITY_NONE  0
#define SERIAL_PARITY_EVEN  1
#define SERIAL_PARITY_ODD   2
#define SERIAL_PARITY_MARK  3
#define SERIAL_PARITY_SPACE 4

struct serial_config {
	int speed;
	int stop_bits;
	int word_size;
	int parity;
};

typedef struct serial serial_t;
typedef struct serial_config serial_config_t;
typedef void (*serial_handler_t)(serial_t *serial, int serial_event);

serial_t *serial_init();
serial_handler_t serial_notification_handler(serial_t *serial, serial_handler_t handler);
int serial_open(serial_t *serial, const char *device, serial_config_t *config);
int serial_close(serial_t *serial);
void serial_release(serial_t *serial);
int serial_read(serial_t *serial, void *data, int length, long timeout);
int serial_write(serial_t *serial, void *data, int length);
void serial_empty_data(serial_t *serial);

serial_config_t *serial_config_create(int speed, const char *options);
#define serial_config_create()	serial_config_create(-1, NULL)
void serial_config_release(serial_config_t *config);

char *serial_device(int port_number, int usb);
int serial_read_bytes_retrys(serial_t * serial, void * buff, int length, int retries, long tout);
int serial_write_bytes_retrys(serial_t * serial, void * buff, int length, int retries);
int serial_write_bytes(serial_t * serial, void * buff, int length);
int serial_read_bytes(serial_t * serial, void * buff, int length);
#endif

