#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include "telmicom.h"

int telmicom_byteToInt(unsigned char *b, int length)
{
	int v = 0, i;
	for(i=length;i; i--) v = (v << 8) + (b[i-1] & 0x000000ff);
	return v;
}

void telmicom_intToByte(unsigned char *b, int v, int length)
{
	int i;
	for(i=0;i < length; i++) {
		b[i] = v & 0x000000ff;
		v >>=  8;
	}
}


unsigned char checksum(const telmicom_data_t* command)
{
	unsigned char cs=0;
	int i;
	cs += command->dst;
	cs += command->org;
	cs += command->cmd;
	cs += command->len;
	for (i = 0; i < command->len; i++)cs += command->data[i];
	return -cs;
}

void calculate_checksum(telmicom_data_t* command)
{
	command->cs = checksum(command);
}

int test_checksum(telmicom_data_t* command)
{
	return command->cs == checksum(command);
}


int _telmicom_sendcommand(serial_t * serial, telmicom_data_t* command, telmicom_data_t* response, int retries, int tout)
{
	unsigned char buff[261];
	int numread=0;
	int numwrote=0;
	int i;
	//print_telmicom_data("Telmicom enviant: ",command);
	calculate_checksum(command);
	buff[0] = command->dst;
	buff[1] = command->len;
	buff[2] = command->org;
	buff[3] = command->cmd;
	memcpy(buff+4,command->data, command->len);
	buff[4+(command->len)] =  command->cs;
	if(serial_write_bytes(serial, buff, (command->len)+5) < (command->len)+5) { /*printf("error a serial_write_bytes\n");*/ goto errio;}
	memset(buff,0,261);
	memset(response,0,sizeof(telmicom_data_t));
	if((numread = serial_read_bytes_retrys(serial, buff, 2, retries, tout)) < 2) { /*printf("error a serial_read 1\n");*/ goto errio;}
	if((numread = serial_read_bytes_retrys(serial, buff+2,buff[1]+3, retries, tout)) < buff[1]+3)  { /*printf("error a serial_read 2\n");*/ goto errio;}
	response->dst = buff[0];
	response->len = buff[1];
	response->org = buff[2];
	response->cmd = buff[3];
	memcpy(response->data,buff+4,buff[1]);
	response->cs = buff[buff[1]+4];
	if(!test_checksum(response)) { /*printf("error a chacksum\n");*/ return TELMICOM_ERR_CHECKSUM;}
	if(command->org != response->dst || command->dst != response->org)
		return TELMICOM_ERR_ADDRESS;
	return 0;

	errio:
//		for (i=0; i < numread; i++) printf(" %x",buff[i]);
//		printf("\n");
		//serial_empty_data(serial);
		return TELMICOM_ERR_IO;
}

int telmicom_sendcommand_retrys(serial_t * serial, telmicom_data_t* command, telmicom_data_t* response, int serial_retries, int serial_tout, int command_retries)
{
	int i;
	for (i=0; i < command_retries; i++) {
		if(!_telmicom_sendcommand(serial,command,response, serial_retries, serial_tout)) return 0;
	}
	return TELMICOM_ERR_IO;
}

int telmicom_sendcommand(serial_t * serial, telmicom_data_t* command, telmicom_data_t* response)
{
	telmicom_sendcommand_retrys(serial,command,response, 1, 200, 5);
}

void print_telmicom_data(const char * label, const telmicom_data_t* command)
{
	int i;
	printf("%s: dst %x len %x org %x cmd %x Dades [",label==NULL?"":label,command->dst, command->len, command->org, command->cmd);
	for (i=0; i < command->len; i++) printf("%02x ",command->data[i]);
	printf("] cs %x (%x)\n",command->cs, checksum(command) );
}

int telmicom_moneda(serial_t * serial, unsigned char org, unsigned char dst)
{
	telmicom_data_t comanda, resposta;
	int res;
	comanda.cmd = 0xff;
	comanda.len = 0x02;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x6a;
	comanda.data[1] = 0x00;
	res = telmicom_sendcommand_retrys(serial, &comanda, &resposta,1,200,5);
	if(res) {
/*		printf("********* Error a telmicom_moneda: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
*/		return res;
	}
	return resposta.data[1];
}

int telmicom_inhabilita(serial_t * serial, unsigned char org, unsigned char dst, int inhabilita)
{
	telmicom_data_t comanda, resposta;
	int i,res;
	char buff[40];
	char errStr[] = " INET ERR";
	char * t;
	comanda.cmd = 0xff;
	comanda.len = 0x01;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x63;
	res = telmicom_sendcommand_retrys(serial, &comanda, &resposta,1,300,5);
	print_telmicom_data("------------------- telmicom_inhabilita_res", &resposta);
	printf("res = %d (%d)\n",res,inhabilita);
	if(res||(resposta.len != 15))return -1;
	comanda.len=16;
	memcpy(comanda.data+1, resposta.data,15);
	comanda.data[0] = 0x65;
	if(inhabilita) {
		comanda.data[15] |= 0x08;
//		comanda.data[15] = 0x08;
	} else {
		comanda.data[15] &= ~0x08;
//		comanda.data[15] = 0x00;
	}
	print_telmicom_data("------------------- telmicom_inhabilita_cmd", &comanda);
	res = telmicom_sendcommand(serial, &comanda, &resposta),1,300,5;
	if(res) return -2;
	res = telmicom_getText(serial, org, dst, buff);
	if(res) return -3;
	printf("------------------- telmicom_inhabilita str: %s\n", buff);
	if(inhabilita) {
		printf("------------------- telmicom_inhabilita 1 str: %s", buff);
		if (((t = strstr(buff,errStr)) == NULL) && (strlen(buff) + strlen(errStr) < 32)) {
			printf("------------------- telmicom_inhabilita 2 str: %s\n", buff);
			strcat(buff,errStr);
		}
	} else {
		printf("------------------- telmicom_inhabilita 3 str: %s\n", buff);
		if (((t = strstr(buff,errStr)) != NULL) && (strlen(t) == strlen(errStr))) {
			printf("------------------- telmicom_inhabilita 4 str: %s\n", buff);
			*t = 0;
		}
	}
	res = telmicom_setText(serial, org, dst, buff);
	printf("------------------- telmicom_inhabilita 5 str: %s res: %d\n", buff, res);
	if(res) return -4;
	return 0;
}

int telmicom_temps(serial_t * serial, unsigned char org, unsigned char dst)
{
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0xff;
	comanda.len = 0x01;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x63;
	res = telmicom_sendcommand_retrys(serial, &comanda, &resposta,1,200,5);
	if(res) {
/*		printf("********* Error a telmicom_temps: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
*/		return res;
	} else {
/*		printf("********* Got telmicom_temps: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
*/	}
	if(resposta.len != 15) {
//		printf("********* Error a telmicom_temps: len %d\n",resposta.len);
		return -99;
	}
	t1 = telmicom_byteToInt(resposta.data,4);
	t2 = telmicom_byteToInt(resposta.data+8,4);
//	printf("t1:%d,t2:%d, t: %d \n", t1, t2, t1+t2);
	return t1+t2;
}

int telmicom_settemps(serial_t * serial, unsigned char org, unsigned char dst, int segons)
{
	printf("telmicom_settemps %d\n",segons);
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0xff;
	comanda.len = 0x10;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x65;
	memset(comanda.data+1,0,15);
	telmicom_intToByte(comanda.data+9,segons,4);
	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_settemps: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
	}
	return 0;
}

double telmicom_get_tarifa(serial_t * serial, unsigned char org, unsigned char dst)
{
	telmicom_data_t comanda, resposta;
	int res;
	comanda.cmd = 0xff;
	comanda.len = 0x01;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x68;
	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_settemps: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
	}
	return ((double)telmicom_byteToInt(resposta.data,2))/100;
}

int telmicom_set_tarifa(serial_t * serial, unsigned char org, unsigned char dst, double tarifa)
{
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0xff;
	comanda.len = 0x03;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x69;
	memset(comanda.data+1,0,15);
	telmicom_intToByte(comanda.data+1,(int)(tarifa*100),2);
	print_telmicom_data("\ttelmicom_set_tarifa Commanda:", &comanda);
	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_set_tarifa: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
	}
	return 0;
}


int telmicom_descuento(serial_t * serial, unsigned char org, unsigned char dst, int activar)
{
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0xff;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x67;
	comanda.len = 2;
	if (activar) comanda.data[1] = 0x04;
	else comanda.data[1] = 0x00;
	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_descuento escriure: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
	}

	return 0;
}



int telmicom_setText(serial_t * serial, unsigned char org, unsigned char dst, const char * text)
{
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0x15;
//	comanda.len = strlen(text) +1;
//	if (comanda.len > 32) comanda.len = 32;
	comanda.len = 32;
	comanda.org = org;
	comanda.dst = dst;
	strncpy(comanda.data,text, 31);
	comanda.data[31]=0;

	res = telmicom_sendcommand_retrys(serial, &comanda, &resposta, 1, 500, 10);
	if(res) {
		printf("********* Error a telmicom_setText escriure: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
	}

	return 0;
}


int telmicom_setFuncionesOciomatik(serial_t * serial, unsigned char org, unsigned char dst,
		int activarReset, int tempsReset, int activarWatchdog, int tempsWatchdog,
		int activarPower, int tempsPower)
{
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0xff;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x6e;
	comanda.len = 8;
	comanda.data[1] = 0;
	if(activarReset) comanda.data[1] |= 1;
	if(activarWatchdog) comanda.data[1] |= 2;
	if(activarPower) comanda.data[1] |= 4;

	telmicom_intToByte(comanda.data + 2,tempsReset,2);
	telmicom_intToByte(comanda.data + 4,tempsPower,2);
	telmicom_intToByte(comanda.data + 6,tempsWatchdog,2);

	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_setFuncionesOciomatik escriure: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
	}

	return 0;
}

int telmicom_setMonedes(serial_t * serial, unsigned char org, unsigned char dst, int monedes[])
{
	telmicom_data_t comanda, resposta;
	int i,res,m=0;
	int monedesMap[] = {0x0010, 0x0040, 0x0200, 0x0400, 0x2000};
	comanda.cmd = 0xe7;
	comanda.len = 0x02;
	comanda.org = org;
	comanda.dst = dst;
	for (i=0; i < 5; i++) if (monedes[i]) m |= monedesMap[i];
	telmicom_intToByte(comanda.data,m,2);
	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_descuento escriure: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
	}

	return 0;
}

int telmicom_getFuncionesOciomatik(serial_t * serial, unsigned char org, unsigned char dst,
		int *activarReset, int *tempsReset, int *activarWatchdog, int *tempsWatchdog,
		int *activarPower, int *tempsPower)
{
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0xff;
	comanda.org = org;
	comanda.dst = dst;
	comanda.data[0] = 0x6d;
	comanda.len = 1;

	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_getFuncionesOciomatik escriure: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
/*	} else {
		printf("------------ Get Funciones Ociomatik: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
*/	}

	*activarReset = (resposta.data[0] & 1)? 1:0;
	*activarWatchdog = (resposta.data[0] & 2)? 1:0;
	*activarPower = (resposta.data[0] & 4)? 1:0;

	*tempsReset = telmicom_byteToInt(resposta.data+1,2);
	*tempsPower = telmicom_byteToInt(resposta.data+3,2);
	*tempsWatchdog = telmicom_byteToInt(resposta.data+5,2);

	return 0;
}

int telmicom_getMonedes(serial_t * serial, unsigned char org, unsigned char dst, int *monedes)
{
	telmicom_data_t comanda, resposta;
	int i,res,m=0;
	int monedesMap[] = {0x0010, 0x0040, 0x0200, 0x0400, 0x2000};
	comanda.cmd = 0xe6;
	comanda.len = 0x0;
	comanda.org = org;
	comanda.dst = dst;

	res = telmicom_sendcommand(serial, &comanda, &resposta);
	if(res) {
		printf("********* Error a telmicom_descuento escriure: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		return res;
/*	} else {
		printf("------------ Get Monedes: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
*/	}

	m = telmicom_byteToInt(resposta.data,2);
	for (i=0; i < 5; i++) monedes[i] = (m & monedesMap[i])? 1:0;
	return 0;
}


int telmicom_getText(serial_t * serial, unsigned char org, unsigned char dst, char * text)
{
	telmicom_data_t comanda, resposta;
	int t1,t2,res;
	comanda.cmd = 0x14;
	comanda.len = 0;
	comanda.org = org;
	comanda.dst = dst;

	res = telmicom_sendcommand_retrys(serial, &comanda, &resposta, 1, 500, 10);
	if(res) {
		printf("********* Error a telmicom_getText escriure: %d\n",res);
		print_telmicom_data("\tCommanda:", &comanda);
		print_telmicom_data("\tResposta:", &resposta);
		if(res != TELMICOM_ERR_CHECKSUM) return res;
	}

	strncpy(text,resposta.data, 31);
	text[31]=0;

	return 0;
}

